
(function () {

    /**
     * 链表节点构造函数
     * @param element
     * @constructor
     */
    let Node = function (element) {
        this.element = element;
        this.next = null;
        this.prev = null;
    };

    /**
     * 链表构造函数
     * @constructor
     */
    let List = function () {
        this.length = 0;
        this.head = null;
        this.tail = null;
    };

    let pro = List.prototype;

    /**
     * 追加元素
     * @param element
     * @returns {List}
     */
    pro.append = function (element) {

        let node = new Node(element);

        if (this.head === null) {
            this.head = node;
            this.tail = node;
        } else {
            this.tail.next = node;
            node.prev = this.tail;
            this.tail = node;
        }

        this.length++;
        return this;
    };

    /**
     * 向指定位置插入元素
     * @param position
     * @param element
     * @returns {boolean}
     */
    pro.insert = function (position, element) {

        if (position >= 0 && position <= this.length) {
            let node = new Node(element),
                current = this.head,
                previous,
                index = 0;

            if (position === 0) {
                if (this.head === null) {
                    this.head = node;
                    this.tail = node;
                } else {
                    node.next = current;
                    current.prev = node;
                    this.head = node;
                }
            } else if (position === this.length) {
                current = this.tail;
                current.next = node;
                node.prev = current;
                this.tail = node;
            } else {
                while (index++ < position) {
                    previous = current;
                    current = current.next;
                }
                node.next = current;
                previous.next = node;

                current.prev = node;
                node.prev = previous;
            }

            this.length++;
            return true;
        } else {
            return false;
        }
    };

    /**
     * 移除指定位置的元素
     * @param position
     * @returns {*}
     */
    pro.removeAt = function (position) {

        if (position > -1 && position < this.length) {
            let current = this.head,
                previous,
                index = 0;

            if (position === 0) {

                this.head = current.next;

                if (this.length === 1) {
                    this.tail = null;
                } else {
                    this.head.prev = null;
                }

            } else if (position === this.length - 1) {
                current = this.tail;
                this.tail = current.prev;
                this.tail.next = null;

            } else {

                while (index++ < position) {

                    previous = current;
                    current = current.next;
                }

                previous.next = current.next;
                current.next.prev = previous;
            }

            this.length--;

            return current.element;

        } else {
            return null;
        }
    };

    /**
     * 移除元素
     * @param element
     * @returns {*}
     */
    pro.remove = function (element) {

        let index = this.indexOf(element);
        return this.removeAt(index);
    };

    /**
     * 指定元素是否存在
     * @param element
     * @returns {number}
     */
    pro.indexOf = function (element) {

        let current = this.head,
            index = -1;

        if (element === current.element) {
            return 0;
        }

        index++;

        while (current.next) {

            if (element === current.element) {
                return index;
            }

            current = current.next;
            index++;
        }

        if (element === current.element) {
            return index;
        }

        return -1;
    };

    /**
     * 链表是否为空
     * @returns {boolean}
     */
    pro.isEmpty = function () {
        return this.length === 0;
    };

    /**
     * 链表的长度
     * @returns {number}
     */
    pro.size = function () {
        return this.length;
    };

    /**
     * 获取链表头元素
     * @returns {null|*}
     */
    pro.getHead = function () {
        return this.head;
    };

    /**
     * 获取链表尾元素
     * @returns {null|*}
     */
    pro.getTail = function () {
        return this.tail;
    };

    /**
     * 对链表中元素进行迭代
     * @param fn
     */
    pro.each = function (fn) {
        let current = this.head;
        while (current) {
            fn(current.element);
            current = current.next;
        }
    };

    /**
     * 清理链表
     */
    pro.clear = function () {
        this.head = null;
        this.tail = null;
        this.length = 0;
    };


    module.exports = List;
    cc.bsList = function () {
        return new List();
    }
})();
